"""
This implementation is NOT complete, there are many things to improve.
The following is a list of the most important missing features and inaccuracies.

TODO add support for more principals, apart from IAM users and assumed IAM roles
TODO add support for the Resource and Condition parts of IAM policies
TODO add support and create tests for all services in moto (for example, API Gateway is probably not supported currently)
TODO implement service specific error messages (currently, EC2 and S3 are supported separately, everything else defaults to the errors IAM returns)
TODO include information about the action's resource in error messages (once the Resource element in IAM policies is supported)
TODO check all other actions that are performed by the action called by the user (for example, autoscaling:CreateAutoScalingGroup requires permission for iam:CreateServiceLinkedRole too - see https://docs.aws.amazon.com/autoscaling/ec2/userguide/control-access-using-iam.html)
TODO add support for resource-based policies

"""

import json
import logging
import re
from abc import abstractmethod, ABCMeta
from enum import Enum
from typing import Any, Dict, Optional, Match, List, Union

from botocore.auth import SigV4Auth, S3SigV4Auth
from botocore.awsrequest import AWSRequest
from botocore.credentials import Credentials

from moto.core.exceptions import (
    SignatureDoesNotMatchError,
    AccessDeniedError,
    InvalidClientTokenIdError,
    AuthFailureError,
)
from moto.s3.exceptions import (
    BucketAccessDeniedError,
    S3AccessDeniedError,
    BucketInvalidTokenError,
    S3InvalidTokenError,
    S3InvalidAccessKeyIdError,
    BucketInvalidAccessKeyIdError,
    BucketSignatureDoesNotMatchError,
    S3SignatureDoesNotMatchError,
)
from moto.sts.models import sts_backends
from .models import iam_backends, Policy, IAMBackend

log = logging.getLogger(__name__)


def create_access_key(
    account_id: str, access_key_id: str, headers: Dict[str, str]
) -> Union["IAMUserAccessKey", "AssumedRoleAccessKey"]:
    if access_key_id.startswith("AKIA") or "X-Amz-Security-Token" not in headers:
        return IAMUserAccessKey(account_id, access_key_id, headers)
    else:
        return AssumedRoleAccessKey(account_id, access_key_id, headers)


class IAMUserAccessKey:
    @property
    def backend(self) -> IAMBackend:
        return iam_backends[self.account_id]["global"]

    def __init__(self, account_id: str, access_key_id: str, headers: Dict[str, str]):
        self.account_id = account_id
        iam_users = self.backend.list_users("/", None, None)

        for iam_user in iam_users:
            for access_key in iam_user.access_keys:
                if access_key.access_key_id == access_key_id:
                    self._owner_user_name = iam_user.name
                    self._access_key_id = access_key_id
                    self._secret_access_key = access_key.secret_access_key
                    if "X-Amz-Security-Token" in headers:
                        raise CreateAccessKeyFailure(reason="InvalidToken")
                    return
        raise CreateAccessKeyFailure(reason="InvalidId")

    @property
    def arn(self) -> str:
        return f"arn:aws:iam::{self.account_id}:user/{self._owner_user_name}"

    def create_credentials(self) -> Credentials:
        return Credentials(self._access_key_id, self._secret_access_key)

    def collect_policies(self) -> List[Dict[str, str]]:
        user_policies = []

        inline_policy_names = self.backend.list_user_policies(self._owner_user_name)
        for inline_policy_name in inline_policy_names:
            inline_policy = self.backend.get_user_policy(
                self._owner_user_name, inline_policy_name
            )
            user_policies.append(inline_policy)

        attached_policies, _ = self.backend.list_attached_user_policies(
            self._owner_user_name
        )
        user_policies += attached_policies

        user_groups = self.backend.get_groups_for_user(self._owner_user_name)
        for user_group in user_groups:
            inline_group_policy_names = self.backend.list_group_policies(
                user_group.name
            )
            for inline_group_policy_name in inline_group_policy_names:
                inline_user_group_policy = self.backend.get_group_policy(
                    user_group.name, inline_group_policy_name
                )
                user_policies.append(inline_user_group_policy)

            attached_group_policies, _ = self.backend.list_attached_group_policies(
                user_group.name
            )
            user_policies += attached_group_policies

        return user_policies


class AssumedRoleAccessKey:
    @property
    def backend(self) -> IAMBackend:
        return iam_backends[self.account_id]["global"]

    def __init__(self, account_id: str, access_key_id: str, headers: Dict[str, str]):
        self.account_id = account_id
        for assumed_role in sts_backends[account_id]["global"].assumed_roles:
            if assumed_role.access_key_id == access_key_id:
                self._access_key_id = access_key_id
                self._secret_access_key = assumed_role.secret_access_key
                self._session_token = assumed_role.session_token
                self._owner_role_name = assumed_role.role_arn.split("/")[-1]
                self._session_name = assumed_role.session_name
                if headers["X-Amz-Security-Token"] != self._session_token:
                    raise CreateAccessKeyFailure(reason="InvalidToken")
                return
        raise CreateAccessKeyFailure(reason="InvalidId")

    @property
    def arn(self) -> str:
        return f"arn:aws:sts::{self.account_id}:assumed-role/{self._owner_role_name}/{self._session_name}"

    def create_credentials(self) -> Credentials:
        return Credentials(
            self._access_key_id, self._secret_access_key, self._session_token
        )

    def collect_policies(self) -> List[str]:
        role_policies = []

        inline_policy_names = self.backend.list_role_policies(self._owner_role_name)
        for inline_policy_name in inline_policy_names:
            _, inline_policy = self.backend.get_role_policy(
                self._owner_role_name, inline_policy_name
            )
            role_policies.append(inline_policy)

        attached_policies, _ = self.backend.list_attached_role_policies(
            self._owner_role_name
        )
        role_policies += attached_policies  # type: ignore[arg-type]

        return role_policies


class CreateAccessKeyFailure(Exception):
    def __init__(self, reason: str):
        super().__init__()
        self.reason = reason


class IAMRequestBase(object, metaclass=ABCMeta):
    def __init__(
        self,
        account_id: str,
        method: str,
        path: str,
        data: Dict[str, str],
        headers: Dict[str, str],
    ):
        log.debug(
            f"Creating {self.__class__.__name__} with method={method}, path={path}, data={data}, headers={headers}"
        )
        self.account_id = account_id
        self._method = method
        self._path = path
        self._data = data
        self._headers = headers
        credential_scope = self._get_string_between(
            "Credential=", ",", self._headers["Authorization"]
        )
        credential_data = credential_scope.split("/")
        self._region = credential_data[2]
        self._service = credential_data[3]
        self._action = (
            self._service
            + ":"
            + (
                self._data["Action"][0]
                if isinstance(self._data["Action"], list)
                else self._data["Action"]
            )
        )
        try:
            self._access_key = create_access_key(
                account_id=self.account_id,
                access_key_id=credential_data[0],
                headers=headers,
            )
        except CreateAccessKeyFailure as e:
            self._raise_invalid_access_key(e.reason)

    def check_signature(self) -> None:
        original_signature = self._get_string_between(
            "Signature=", ",", self._headers["Authorization"]
        )
        calculated_signature = self._calculate_signature()
        if original_signature != calculated_signature:
            self._raise_signature_does_not_match()

    def check_action_permitted(self, resource: str) -> None:
        if (
            self._action == "sts:GetCallerIdentity"
        ):  # always allowed, even if there's an explicit Deny for it
            return
        policies = self._access_key.collect_policies()

        permitted = False
        for policy in policies:
            iam_policy = IAMPolicy(policy)
            permission_result = iam_policy.is_action_permitted(self._action, resource)
            if permission_result == PermissionResult.DENIED:
                self._raise_access_denied()
            elif permission_result == PermissionResult.PERMITTED:
                permitted = True

        if not permitted:
            self._raise_access_denied()

    @abstractmethod
    def _raise_signature_does_not_match(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    def _raise_access_denied(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    def _raise_invalid_access_key(self, reason: str) -> None:
        raise NotImplementedError()

    @abstractmethod
    def _create_auth(self, credentials: Credentials) -> SigV4Auth:  # type: ignore[misc]
        raise NotImplementedError()

    @staticmethod
    def _create_headers_for_aws_request(
        signed_headers: List[str], original_headers: Dict[str, str]
    ) -> Dict[str, str]:
        headers = {}
        for key, value in original_headers.items():
            if key.lower() in signed_headers:
                headers[key] = value
        return headers

    def _create_aws_request(self) -> AWSRequest:
        signed_headers = self._get_string_between(
            "SignedHeaders=", ",", self._headers["Authorization"]
        ).split(";")
        headers = self._create_headers_for_aws_request(signed_headers, self._headers)
        request = AWSRequest(
            method=self._method, url=self._path, data=self._data, headers=headers
        )
        request.context["timestamp"] = headers["X-Amz-Date"]

        return request

    def _calculate_signature(self) -> str:
        credentials = self._access_key.create_credentials()
        auth = self._create_auth(credentials)
        request = self._create_aws_request()
        canonical_request = auth.canonical_request(request)
        string_to_sign = auth.string_to_sign(request, canonical_request)
        return auth.signature(string_to_sign, request)

    @staticmethod
    def _get_string_between(
        first_separator: str, second_separator: str, string: str
    ) -> str:
        return string.partition(first_separator)[2].partition(second_separator)[0]


class IAMRequest(IAMRequestBase):
    def _raise_signature_does_not_match(self) -> None:
        if self._service == "ec2":
            raise AuthFailureError()
        else:
            raise SignatureDoesNotMatchError()

    def _raise_invalid_access_key(self, _: str) -> None:
        if self._service == "ec2":
            raise AuthFailureError()
        else:
            raise InvalidClientTokenIdError()

    def _create_auth(self, credentials: Any) -> SigV4Auth:
        return SigV4Auth(credentials, self._service, self._region)

    def _raise_access_denied(self) -> None:
        raise AccessDeniedError(user_arn=self._access_key.arn, action=self._action)


class S3IAMRequest(IAMRequestBase):
    def _raise_signature_does_not_match(self) -> None:
        if "BucketName" in self._data:
            raise BucketSignatureDoesNotMatchError(bucket=self._data["BucketName"])
        else:
            raise S3SignatureDoesNotMatchError()

    def _raise_invalid_access_key(self, reason: str) -> None:
        if reason == "InvalidToken":
            if "BucketName" in self._data:
                raise BucketInvalidTokenError(bucket=self._data["BucketName"])
            else:
                raise S3InvalidTokenError()
        else:
            if "BucketName" in self._data:
                raise BucketInvalidAccessKeyIdError(bucket=self._data["BucketName"])
            else:
                raise S3InvalidAccessKeyIdError()

    def _create_auth(self, credentials: Any) -> S3SigV4Auth:
        return S3SigV4Auth(credentials, self._service, self._region)

    def _raise_access_denied(self) -> None:
        if "BucketName" in self._data:
            raise BucketAccessDeniedError(bucket=self._data["BucketName"])
        else:
            raise S3AccessDeniedError()


class IAMPolicy:
    def __init__(self, policy: Any):
        if isinstance(policy, Policy):
            default_version = next(
                policy_version
                for policy_version in policy.versions
                if policy_version.is_default
            )
            policy_document = default_version.document
        elif isinstance(policy, str):
            policy_document = policy
        else:
            policy_document = policy["policy_document"]

        self._policy_json = json.loads(policy_document)

    def is_action_permitted(
        self, action: str, resource: str = "*"
    ) -> "PermissionResult":
        permitted = False
        if isinstance(self._policy_json["Statement"], list):
            for policy_statement in self._policy_json["Statement"]:
                iam_policy_statement = IAMPolicyStatement(policy_statement)
                permission_result = iam_policy_statement.is_action_permitted(
                    action, resource
                )
                if permission_result == PermissionResult.DENIED:
                    return permission_result
                elif permission_result == PermissionResult.PERMITTED:
                    permitted = True
        else:  # dict
            iam_policy_statement = IAMPolicyStatement(self._policy_json["Statement"])
            return iam_policy_statement.is_action_permitted(action)

        if permitted:
            return PermissionResult.PERMITTED
        else:
            return PermissionResult.NEUTRAL


class IAMPolicyStatement:
    def __init__(self, statement: Any):
        self._statement = statement

    def is_action_permitted(
        self, action: str, resource: str = "*"
    ) -> "PermissionResult":
        is_action_concerned = False

        if "NotAction" in self._statement:
            if not self._check_element_matches("NotAction", action):
                is_action_concerned = True
        else:  # Action is present
            if self._check_element_matches("Action", action):
                is_action_concerned = True

        if is_action_concerned:
            if self.is_unknown_principal(self._statement.get("Principal")):
                return PermissionResult.NEUTRAL
            same_resource = self._check_element_matches("Resource", resource)
            if self._statement["Effect"] == "Allow" and same_resource:
                return PermissionResult.PERMITTED
            else:  # Deny
                return PermissionResult.DENIED
        else:
            return PermissionResult.NEUTRAL

    def is_unknown_principal(self, principal: Optional[str]) -> bool:
        # https://docs.aws.amazon.com/AmazonS3/latest/userguide/s3-bucket-user-policy-specifying-principal-intro.html
        # For now, Moto only verifies principal == *
        # 'Unknown' principals are not verified
        #
        # This should be extended to check:
        # - Can the principal be empty? How behaves AWS?
        # - allow one/multiple account ARN's
        # - allow one/multiple rules
        if principal is None:
            return False
        if isinstance(principal, str) and principal != "*":
            return True
        return False

    def _check_element_matches(self, statement_element: Any, value: str) -> bool:
        if isinstance(self._statement[statement_element], list):
            for statement_element_value in self._statement[statement_element]:
                if self._match(statement_element_value, value):
                    return True
            return False
        else:  # string
            return self._match(self._statement[statement_element], value) is not None

    @staticmethod
    def _match(pattern: str, string: str) -> Optional[Match[str]]:
        pattern = pattern.replace("*", ".*")
        pattern = f"^{pattern}$"
        return re.match(pattern, string)


class PermissionResult(Enum):
    PERMITTED = 1
    DENIED = 2
    NEUTRAL = 3



from moto.core.responses import BaseResponse
from .exceptions import PasswordTooShort, PasswordRequired
from .models import elasticache_backends, ElastiCacheBackend


class ElastiCacheResponse(BaseResponse):
    """Handler for ElastiCache requests and responses."""

    def __init__(self) -> None:
        super().__init__(service_name="elasticache")

    @property
    def elasticache_backend(self) -> ElastiCacheBackend:
        """Return backend instance specific for this region."""
        return elasticache_backends[self.current_account][self.region]

    def create_user(self) -> str:
        params = self._get_params()
        user_id = params.get("UserId")
        user_name = params.get("UserName")
        engine = params.get("Engine")
        passwords = params.get("Passwords", [])
        no_password_required = self._get_bool_param("NoPasswordRequired", False)
        password_required = not no_password_required
        if password_required and not passwords:
            raise PasswordRequired
        if any([len(p) < 16 for p in passwords]):
            raise PasswordTooShort
        access_string = params.get("AccessString")
        user = self.elasticache_backend.create_user(
            user_id=user_id,  # type: ignore[arg-type]
            user_name=user_name,  # type: ignore[arg-type]
            engine=engine,  # type: ignore[arg-type]
            passwords=passwords,
            access_string=access_string,  # type: ignore[arg-type]
            no_password_required=no_password_required,
        )
        template = self.response_template(CREATE_USER_TEMPLATE)
        return template.render(user=user)

    def delete_user(self) -> str:
        params = self._get_params()
        user_id = params.get("UserId")
        user = self.elasticache_backend.delete_user(user_id=user_id)  # type: ignore[arg-type]
        template = self.response_template(DELETE_USER_TEMPLATE)
        return template.render(user=user)

    def describe_users(self) -> str:
        params = self._get_params()
        user_id = params.get("UserId")
        users = self.elasticache_backend.describe_users(user_id=user_id)
        template = self.response_template(DESCRIBE_USERS_TEMPLATE)
        return template.render(users=users)


USER_TEMPLATE = """<UserId>{{ user.id }}</UserId>
    <UserName>{{ user.name }}</UserName>
    <Status>{{ user.status }}</Status>
    <Engine>{{ user.engine }}</Engine>
    <MinimumEngineVersion>{{ user.minimum_engine_version }}</MinimumEngineVersion>
    <AccessString>{{ user.access_string }}</AccessString>
    <UserGroupIds>
{% for usergroupid in user.usergroupids %}
      <member>{{ usergroupid }}</member>
{% endfor %}
    </UserGroupIds>
    <Authentication>
      {% if user.no_password_required %}
      <Type>no-password</Type>
      {% else %}
      <Type>password</Type>
      <PasswordCount>{{ user.passwords|length }}</PasswordCount>
      {% endif %}
    </Authentication>
    <ARN>{{ user.arn }}</ARN>"""


CREATE_USER_TEMPLATE = (
    """<CreateUserResponse xmlns="http://elasticache.amazonaws.com/doc/2015-02-02/">
  <ResponseMetadata>
    <RequestId>1549581b-12b7-11e3-895e-1334aEXAMPLE</RequestId>
  </ResponseMetadata>
  <CreateUserResult>
    """
    + USER_TEMPLATE
    + """
  </CreateUserResult>
</CreateUserResponse>"""
)

DELETE_USER_TEMPLATE = (
    """<DeleteUserResponse xmlns="http://elasticache.amazonaws.com/doc/2015-02-02/">
  <ResponseMetadata>
    <RequestId>1549581b-12b7-11e3-895e-1334aEXAMPLE</RequestId>
  </ResponseMetadata>
  <DeleteUserResult>
    """
    + USER_TEMPLATE
    + """
  </DeleteUserResult>
</DeleteUserResponse>"""
)

DESCRIBE_USERS_TEMPLATE = (
    """<DescribeUsersResponse xmlns="http://elasticache.amazonaws.com/doc/2015-02-02/">
  <ResponseMetadata>
    <RequestId>1549581b-12b7-11e3-895e-1334aEXAMPLE</RequestId>
  </ResponseMetadata>
  <DescribeUsersResult>
    <Users>
{% for user in users %}
      <member>
        """
    + USER_TEMPLATE
    + """
      </member>
{% endfor %}
    </Users>
    <Marker></Marker>
  </DescribeUsersResult>
</DescribeUsersResponse>"""
)



import json
from typing import Any, Dict, Tuple

from moto.core.exceptions import AWSError


class EKSError(AWSError):
    def __init__(self, **kwargs: Any):
        super(AWSError, self).__init__(error_type=self.TYPE, message="")  # type: ignore
        self.description = json.dumps(kwargs)
        self.headers = {"status": self.STATUS, "x-amzn-ErrorType": self.TYPE}
        self.code = self.STATUS

    def response(self) -> Tuple[int, Dict[str, Any], str]:  # type: ignore[override]
        return self.STATUS, self.headers, self.description


class ResourceInUseException(EKSError):
    TYPE = "ResourceInUseException"
    STATUS = 409


class ResourceNotFoundException(EKSError):
    TYPE = "ResourceNotFoundException"
    STATUS = 404


class InvalidParameterException(EKSError):
    TYPE = "InvalidParameterException"
    STATUS = 400


class InvalidRequestException(EKSError):
    TYPE = "InvalidRequestException"
    STATUS = 400



import json

from moto.core.responses import BaseResponse
from .models import athena_backends, AthenaBackend
from typing import Dict, Tuple, Union


class AthenaResponse(BaseResponse):
    def __init__(self) -> None:
        super().__init__(service_name="athena")

    @property
    def athena_backend(self) -> AthenaBackend:
        return athena_backends[self.current_account][self.region]

    def create_work_group(self) -> Union[Tuple[str, Dict[str, int]], str]:
        name = self._get_param("Name")
        description = self._get_param("Description")
        configuration = self._get_param("Configuration")
        tags = self._get_param("Tags")
        work_group = self.athena_backend.create_work_group(
            name, configuration, description, tags
        )
        if not work_group:
            return self.error("WorkGroup already exists", 400)
        return json.dumps(
            {
                "CreateWorkGroupResponse": {
                    "ResponseMetadata": {
                        "RequestId": "384ac68d-3775-11df-8963-01868b7c937a"
                    }
                }
            }
        )

    def list_work_groups(self) -> str:
        return json.dumps({"WorkGroups": self.athena_backend.list_work_groups()})

    def get_work_group(self) -> str:
        name = self._get_param("WorkGroup")
        return json.dumps({"WorkGroup": self.athena_backend.get_work_group(name)})

    def start_query_execution(self) -> Union[Tuple[str, Dict[str, int]], str]:
        query = self._get_param("QueryString")
        context = self._get_param("QueryExecutionContext")
        config = self._get_param("ResultConfiguration")
        workgroup = self._get_param("WorkGroup")
        if workgroup and not self.athena_backend.get_work_group(workgroup):
            return self.error("WorkGroup does not exist", 400)
        q_exec_id = self.athena_backend.start_query_execution(
            query=query, context=context, config=config, workgroup=workgroup
        )
        return json.dumps({"QueryExecutionId": q_exec_id})

    def get_query_execution(self) -> str:
        exec_id = self._get_param("QueryExecutionId")
        execution = self.athena_backend.get_query_execution(exec_id)
        result = {
            "QueryExecution": {
                "QueryExecutionId": exec_id,
                "Query": execution.query,
                "StatementType": "DDL",
                "ResultConfiguration": execution.config,
                "QueryExecutionContext": execution.context,
                "Status": {
                    "State": execution.status,
                    "SubmissionDateTime": execution.start_time,
                },
                "Statistics": {
                    "EngineExecutionTimeInMillis": 0,
                    "DataScannedInBytes": 0,
                    "TotalExecutionTimeInMillis": 0,
                    "QueryQueueTimeInMillis": 0,
                    "QueryPlanningTimeInMillis": 0,
                    "ServiceProcessingTimeInMillis": 0,
                },
                "WorkGroup": execution.workgroup,
            }
        }
        return json.dumps(result)

    def get_query_results(self) -> str:
        exec_id = self._get_param("QueryExecutionId")
        result = self.athena_backend.get_query_results(exec_id)
        return json.dumps(result.to_dict())

    def list_query_executions(self) -> str:
        executions = self.athena_backend.list_query_executions()
        return json.dumps({"QueryExecutionIds": [i for i in executions.keys()]})

    def stop_query_execution(self) -> str:
        exec_id = self._get_param("QueryExecutionId")
        self.athena_backend.stop_query_execution(exec_id)
        return json.dumps({})

    def error(self, msg: str, status: int) -> Tuple[str, Dict[str, int]]:
        return (
            json.dumps({"__type": "InvalidRequestException", "Message": msg}),
            dict(status=status),
        )

    def create_named_query(self) -> Union[Tuple[str, Dict[str, int]], str]:
        name = self._get_param("Name")
        description = self._get_param("Description")
        database = self._get_param("Database")
        query_string = self._get_param("QueryString")
        workgroup = self._get_param("WorkGroup") or "primary"
        if not self.athena_backend.get_work_group(workgroup):
            return self.error("WorkGroup does not exist", 400)
        query_id = self.athena_backend.create_named_query(
            name, description, database, query_string, workgroup
        )
        return json.dumps({"NamedQueryId": query_id})

    def get_named_query(self) -> str:
        query_id = self._get_param("NamedQueryId")
        nq = self.athena_backend.get_named_query(query_id)
        return json.dumps(
            {
                "NamedQuery": {
                    "Name": nq.name,  # type: ignore[union-attr]
                    "Description": nq.description,  # type: ignore[union-attr]
                    "Database": nq.database,  # type: ignore[union-attr]
                    "QueryString": nq.query_string,  # type: ignore[union-attr]
                    "NamedQueryId": nq.id,  # type: ignore[union-attr]
                    "WorkGroup": nq.workgroup.name,  # type: ignore[union-attr]
                }
            }
        )

    def list_data_catalogs(self) -> str:
        return json.dumps(
            {"DataCatalogsSummary": self.athena_backend.list_data_catalogs()}
        )

    def get_data_catalog(self) -> str:
        name = self._get_param("Name")
        return json.dumps({"DataCatalog": self.athena_backend.get_data_catalog(name)})

    def create_data_catalog(self) -> Union[Tuple[str, Dict[str, int]], str]:
        name = self._get_param("Name")
        catalog_type = self._get_param("Type")
        description = self._get_param("Description")
        parameters = self._get_param("Parameters")
        tags = self._get_param("Tags")
        data_catalog = self.athena_backend.create_data_catalog(
            name, catalog_type, description, parameters, tags
        )
        if not data_catalog:
            return self.error("DataCatalog already exists", 400)
        return json.dumps(
            {
                "CreateDataCatalogResponse": {
                    "ResponseMetadata": {
                        "RequestId": "384ac68d-3775-11df-8963-01868b7c937a"
                    }
                }
            }
        )

    def list_named_queries(self) -> str:
        next_token = self._get_param("NextToken")
        max_results = self._get_param("MaxResults")
        work_group = self._get_param("WorkGroup") or "primary"
        named_query_ids, next_token = self.athena_backend.list_named_queries(
            next_token=next_token, max_results=max_results, work_group=work_group
        )
        return json.dumps({"NamedQueryIds": named_query_ids, "NextToken": next_token})

    def create_prepared_statement(self) -> Union[str, Tuple[str, Dict[str, int]]]:
        statement_name = self._get_param("StatementName")
        work_group = self._get_param("WorkGroup")
        query_statement = self._get_param("QueryStatement")
        description = self._get_param("Description")
        if not self.athena_backend.get_work_group(work_group):
            return self.error("WorkGroup does not exist", 400)
        self.athena_backend.create_prepared_statement(
            statement_name=statement_name,
            workgroup=work_group,
            query_statement=query_statement,
            description=description,
        )
        return json.dumps(dict())

    def get_prepared_statement(self) -> str:
        statement_name = self._get_param("StatementName")
        work_group = self._get_param("WorkGroup")
        ps = self.athena_backend.get_prepared_statement(
            statement_name=statement_name,
            work_group=work_group,
        )
        return json.dumps(
            {
                "PreparedStatement": {
                    "StatementName": ps.statement_name,  # type: ignore[union-attr]
                    "QueryStatement": ps.query_statement,  # type: ignore[union-attr]
                    "WorkGroupName": ps.workgroup,  # type: ignore[union-attr]
                    "Description": ps.description,  # type: ignore[union-attr]
                    # "LastModifiedTime": ps.last_modified_time,  # type: ignore[union-attr]
                }
            }
        )



"""Exceptions raised by the opensearch service."""
from moto.core.exceptions import JsonRESTError


class ResourceNotFoundException(JsonRESTError):
    def __init__(self, name: str):
        super().__init__("ResourceNotFoundException", f"Domain not found: {name}")



from moto.core.exceptions import RESTError
from typing import Any, List, Optional, Union, Iterable


# EC2 has a custom root-tag - <Response> vs <ErrorResponse>
# `terraform destroy` will complain if the roottag is incorrect
# See https://docs.aws.amazon.com/AWSEC2/latest/APIReference/errors-overview.html#api-error-response
EC2_ERROR_RESPONSE = """<?xml version="1.0" encoding="UTF-8"?>
<Response>
  <Errors>
    <Error>
      <Code>{{error_type}}</Code>
      <Message>{{message}}</Message>
    </Error>
  </Errors>
  <{{request_id_tag}}>7a62c49f-347e-4fc4-9331-6e8eEXAMPLE</{{request_id_tag}}>
</Response>
"""


class EC2ClientError(RESTError):
    code = 400
    # EC2 uses <RequestID> as tag name in the XML response
    request_id_tag_name = "RequestID"

    def __init__(self, *args: Any, **kwargs: Any):
        kwargs.setdefault("template", "custom_response")
        self.templates["custom_response"] = EC2_ERROR_RESPONSE
        super().__init__(*args, **kwargs)


class DefaultVpcAlreadyExists(EC2ClientError):
    def __init__(self) -> None:
        super().__init__(
            "DefaultVpcAlreadyExists",
            "A Default VPC already exists for this account in this region.",
        )


class DependencyViolationError(EC2ClientError):
    def __init__(self, message: str):
        super().__init__("DependencyViolation", message)


class MissingParameterError(EC2ClientError):
    def __init__(self, parameter: str):
        super().__init__(
            "MissingParameter",
            f"The request must contain the parameter {parameter}",
        )


class InvalidDHCPOptionsIdError(EC2ClientError):
    def __init__(self, dhcp_options_id: str):
        super().__init__(
            "InvalidDhcpOptionID.NotFound",
            f"DhcpOptionID {dhcp_options_id} does not exist.",
        )


class InvalidRequest(EC2ClientError):
    def __init__(self) -> None:
        super().__init__("InvalidRequest", "The request received was invalid")


class InvalidParameterCombination(EC2ClientError):
    def __init__(self, msg: str):
        super().__init__("InvalidParameterCombination", msg)


class MalformedDHCPOptionsIdError(EC2ClientError):
    def __init__(self, dhcp_options_id: Optional[str]):
        super().__init__(
            "InvalidDhcpOptionsId.Malformed",
            f'Invalid id: "{dhcp_options_id}" (expecting "dopt-...")',
        )


class InvalidKeyPairNameError(EC2ClientError):
    def __init__(self, key: Iterable[str]):
        super().__init__(
            "InvalidKeyPair.NotFound", f"The keypair '{key}' does not exist."
        )


class InvalidKeyPairDuplicateError(EC2ClientError):
    def __init__(self, key: str):
        super().__init__(
            "InvalidKeyPair.Duplicate", f"The keypair '{key}' already exists."
        )


class InvalidKeyPairFormatError(EC2ClientError):
    def __init__(self) -> None:
        super().__init__(
            "InvalidKeyPair.Format", "Key is not in valid OpenSSH public key format"
        )


class InvalidVPCIdError(EC2ClientError):
    def __init__(self, vpc_id: Any):
        super().__init__("InvalidVpcID.NotFound", f"VpcID {vpc_id} does not exist.")


class InvalidSubnetIdError(EC2ClientError):
    def __init__(self, subnet_id: str):
        super().__init__(
            "InvalidSubnetID.NotFound", f"The subnet ID '{subnet_id}' does not exist"
        )


class InvalidFlowLogIdError(EC2ClientError):
    def __init__(self, count: int, flow_log_ids: str):
        super().__init__(
            "InvalidFlowLogId.NotFound",
            f"These flow log ids in the input list are not found: [TotalCount: {count}] {flow_log_ids}",
        )


class FlowLogAlreadyExists(EC2ClientError):
    def __init__(self) -> None:
        super().__init__(
            "FlowLogAlreadyExists",
            "Error. There is an existing Flow Log with the same configuration and log destination.",
        )


class InvalidNetworkAclIdError(EC2ClientError):
    def __init__(self, network_acl_id: str):
        super().__init__(
            "InvalidNetworkAclID.NotFound",
            f"The network acl ID '{network_acl_id}' does not exist",
        )


class InvalidVpnGatewayIdError(EC2ClientError):
    def __init__(self, vpn_gw: str):
        super().__init__(
            "InvalidVpnGatewayID.NotFound",
            f"The virtual private gateway ID '{vpn_gw}' does not exist",
        )


class InvalidVpnGatewayAttachmentError(EC2ClientError):
    def __init__(self, vpn_gw: str, vpc_id: str):
        super().__init__(
            "InvalidVpnGatewayAttachment.NotFound",
            f"The attachment with vpn gateway ID '{vpn_gw}' and vpc ID '{vpc_id}' does not exist",
        )


class InvalidVpnConnectionIdError(EC2ClientError):
    def __init__(self, network_acl_id: str):
        super().__init__(
            "InvalidVpnConnectionID.NotFound",
            f"The vpnConnection ID '{network_acl_id}' does not exist",
        )


class InvalidCustomerGatewayIdError(EC2ClientError):
    def __init__(self, customer_gateway_id: str):
        super().__init__(
            "InvalidCustomerGatewayID.NotFound",
            f"The customer gateway ID '{customer_gateway_id}' does not exist",
        )


class InvalidNetworkInterfaceIdError(EC2ClientError):
    def __init__(self, eni_id: str):
        super().__init__(
            "InvalidNetworkInterfaceID.NotFound",
            f"The network interface ID '{eni_id}' does not exist",
        )


class InvalidNetworkAttachmentIdError(EC2ClientError):
    def __init__(self, attachment_id: str):
        super().__init__(
            "InvalidAttachmentID.NotFound",
            f"The network interface attachment ID '{attachment_id}' does not exist",
        )


class InvalidSecurityGroupDuplicateError(EC2ClientError):
    def __init__(self, name: str):
        super().__init__(
            "InvalidGroup.Duplicate", f"The security group '{name}' already exists"
        )


class InvalidSecurityGroupNotFoundError(EC2ClientError):
    def __init__(self, name: Any):
        super().__init__(
            "InvalidGroup.NotFound",
            f"The security group '{name}' does not exist",
        )


class InvalidPermissionNotFoundError(EC2ClientError):
    def __init__(self) -> None:
        super().__init__(
            "InvalidPermission.NotFound",
            "The specified rule does not exist in this security group",
        )


class InvalidPermissionDuplicateError(EC2ClientError):
    def __init__(self) -> None:
        super().__init__(
            "InvalidPermission.Duplicate", "The specified rule already exists"
        )


class InvalidRouteTableIdError(EC2ClientError):
    def __init__(self, route_table_id: str):
        super().__init__(
            "InvalidRouteTableID.NotFound",
            f"The routeTable ID '{route_table_id}' does not exist",
        )


class InvalidRouteError(EC2ClientError):
    def __init__(self, route_table_id: str, cidr: str):
        super().__init__(
            "InvalidRoute.NotFound",
            f"no route with destination-cidr-block {cidr} in route table {route_table_id}",
        )


class RouteAlreadyExistsError(EC2ClientError):
    def __init__(self, cidr: str):
        super().__init__(
            "RouteAlreadyExists", f"The route identified by {cidr} already exists"
        )


class RouteNotSupportedError(EC2ClientError):
    def __init__(self, vpce_id: str):
        super().__init__(
            "RouteNotSupported",
            f"Route table contains unsupported route target: {vpce_id}. "
            f"VPC Endpoints of this type cannot be used as route targets.",
        )


class InvalidInstanceIdError(EC2ClientError):
    def __init__(self, instance_id: Any):
        if isinstance(instance_id, str):
            instance_id = [instance_id]
        if len(instance_id) > 1:
            msg = f"The instance IDs '{', '.join(instance_id)}' do not exist"
        else:
            msg = f"The instance ID '{instance_id[0]}' does not exist"
        super().__init__("InvalidInstanceID.NotFound", msg)


class InvalidInstanceTypeError(EC2ClientError):
    def __init__(
        self, instance_type: Any, error_type: str = "InvalidInstanceType.NotFound"
    ):
        if isinstance(instance_type, str):
            msg = f"The instance type '{instance_type}' does not exist"
        else:
            msg = (
                f"The following supplied instance types do not exist: "
                f"[{', '.join(instance_type)}]"
            )
        super().__init__(error_type, msg)


class InvalidAMIIdError(EC2ClientError):
    def __init__(self, ami_id: Union[List[str], str]):
        super().__init__(
            "InvalidAMIID.NotFound",
            f"The image id '[{ami_id}]' does not exist",
        )


class UnvailableAMIIdError(EC2ClientError):
    def __init__(self, ami_id: str):
        super().__init__(
            "InvalidAMIID.Unavailable",
            f"The image id '[{ami_id}]' is no longer available",
        )


class InvalidAMIAttributeItemValueError(EC2ClientError):
    def __init__(self, attribute: str, value: Union[str, Iterable[str], None]):
        super().__init__(
            "InvalidAMIAttributeItemValue",
            f'Invalid attribute item value "{value}" for {attribute} item type.',
        )


class MalformedAMIIdError(EC2ClientError):
    def __init__(self, ami_id: List[str]):
        super().__init__(
            "InvalidAMIID.Malformed", f'Invalid id: "{ami_id}" (expecting "ami-...")'
        )


class InvalidSnapshotIdError(EC2ClientError):
    def __init__(self) -> None:
        # Note: AWS returns empty message for this, as of 2014.08.22.
        super().__init__("InvalidSnapshot.NotFound", "")


class InvalidSnapshotInUse(EC2ClientError):
    def __init__(self, snapshot_id: str, ami_id: str):
        super().__init__(
            "InvalidSnapshot.InUse",
            f"The snapshot {snapshot_id} is currently in use by {ami_id}",
        )


class InvalidVolumeIdError(EC2ClientError):
    def __init__(self, volume_id: Any):
        super().__init__(
            "InvalidVolume.NotFound", f"The volume '{volume_id}' does not exist."
        )


class InvalidVolumeAttachmentError(EC2ClientError):
    def __init__(self, volume_id: str, instance_id: str):
        super().__init__(
            "InvalidAttachment.NotFound",
            f"Volume {volume_id} can not be detached from {instance_id} because it is not attached",
        )


class InvalidVolumeDetachmentError(EC2ClientError):
    def __init__(self, volume_id: str, instance_id: str, device: str):
        super().__init__(
            "InvalidAttachment.NotFound",
            f"The volume {volume_id} is not attached to instance {instance_id} as device {device}",
        )


class VolumeInUseError(EC2ClientError):
    def __init__(self, volume_id: str, instance_id: str):
        super().__init__(
            "VolumeInUse",
            f"Volume {volume_id} is currently attached to {instance_id}",
        )


class InvalidAddressError(EC2ClientError):
    def __init__(self, ip: Any):
        super().__init__("InvalidAddress.NotFound", f"Address '{ip}' not found.")


class InvalidAllocationIdError(EC2ClientError):
    def __init__(self, allocation_id: Any):
        super().__init__(
            "InvalidAllocationID.NotFound",
            f"Allocation ID '{allocation_id}' not found.",
        )


class InvalidAssociationIdError(EC2ClientError):
    def __init__(self, association_id: Any):
        super().__init__(
            "InvalidAssociationID.NotFound",
            f"Association ID '{association_id}' not found.",
        )


class InvalidVpcCidrBlockAssociationIdError(EC2ClientError):
    def __init__(self, association_id: str):
        super().__init__(
            "InvalidVpcCidrBlockAssociationIdError.NotFound",
            f"The vpc CIDR block association ID '{association_id}' does not exist",
        )


class InvalidVPCPeeringConnectionIdError(EC2ClientError):
    def __init__(self, vpc_peering_connection_id: str):
        super().__init__(
            "InvalidVpcPeeringConnectionId.NotFound",
            f"VpcPeeringConnectionID {vpc_peering_connection_id} does not exist.",
        )


class InvalidVPCPeeringConnectionStateTransitionError(EC2ClientError):
    def __init__(self, vpc_peering_connection_id: str):
        super().__init__(
            "InvalidStateTransition",
            f"VpcPeeringConnectionID {vpc_peering_connection_id} is not in the correct state for the request.",
        )


class InvalidServiceName(EC2ClientError):
    def __init__(self, service_name: str):
        super().__init__(
            "InvalidServiceName",
            f"The Vpc Endpoint Service '{service_name}' does not exist",
        )


class InvalidFilter(EC2ClientError):
    def __init__(self, filter_name: str, error_type: str = "InvalidFilter"):
        super().__init__(error_type, f"The filter '{filter_name}' is invalid")


class InvalidNextToken(EC2ClientError):
    def __init__(self, next_token: str):
        super().__init__("InvalidNextToken", f"The token '{next_token}' is invalid")


class InvalidDependantParameterError(EC2ClientError):
    def __init__(self, dependant_parameter: str, parameter: str, parameter_value: str):
        super().__init__(
            "InvalidParameter",
            f"{dependant_parameter} can't be empty if {parameter} is {parameter_value}.",
        )


class InvalidDependantParameterTypeError(EC2ClientError):
    def __init__(self, dependant_parameter: str, parameter_value: str, parameter: str):
        super().__init__(
            "InvalidParameter",
            f"{dependant_parameter} type must be {parameter_value} if {parameter} is provided.",
        )


class InvalidAggregationIntervalParameterError(EC2ClientError):
    def __init__(self, parameter: str):
        super().__init__("InvalidParameter", f"Invalid {parameter}")


class InvalidParameterValueError(EC2ClientError):
    def __init__(self, parameter_value: str):
        super().__init__(
            "InvalidParameterValue",
            f"Value {parameter_value} is invalid for parameter.",
        )


class InvalidParameterValueErrorTagSpotFleetRequest(EC2ClientError):
    def __init__(self, resource_type: str):
        super().__init__(
            "InvalidParameterValue",
            f"The value for `ResourceType` must be `spot-fleet-request`, but got `{resource_type}` instead.",
        )


class InvalidParameterValueErrorReplaceRoute(EC2ClientError):
    def __init__(self, cidr: str):
        super().__init__(
            "InvalidParameterValue",
            f"There is no route defined for '{cidr}' in the route table. Use CreateRoute instead.",
        )


class EmptyTagSpecError(EC2ClientError):
    def __init__(self) -> None:
        super().__init__(
            "InvalidParameterValue", "Tag specification must have at least one tag"
        )


class InvalidParameterValueErrorTagNull(EC2ClientError):
    def __init__(self) -> None:
        super().__init__(
            "InvalidParameterValue",
            "Tag value cannot be null. Use empty string instead.",
        )


class InvalidParameterValueErrorUnknownAttribute(EC2ClientError):
    def __init__(self, parameter_value: str):
        super().__init__(
            "InvalidParameterValue",
            f"Value ({parameter_value}) for parameter attribute is invalid. Unknown attribute.",
        )


class InvalidGatewayIDError(EC2ClientError):
    def __init__(self, gateway_id: str):
        super().__init__(
            "InvalidGatewayID.NotFound", f"The eigw ID '{gateway_id}' does not exist"
        )


class InvalidInternetGatewayIdError(EC2ClientError):
    def __init__(self, internet_gateway_id: str):
        super().__init__(
            "InvalidInternetGatewayID.NotFound",
            f"InternetGatewayID {internet_gateway_id} does not exist.",
        )


class GatewayNotAttachedError(EC2ClientError):
    def __init__(self, internet_gateway_id: str, vpc_id: str):
        super().__init__(
            "Gateway.NotAttached",
            f"InternetGatewayID {internet_gateway_id} is not attached to a VPC {vpc_id}.",
        )


class ResourceAlreadyAssociatedError(EC2ClientError):
    def __init__(self, resource_id: str):
        super().__init__(
            "Resource.AlreadyAssociated",
            f"Resource {resource_id} is already associated.",
        )


class TagLimitExceeded(EC2ClientError):
    def __init__(self) -> None:
        super().__init__(
            "TagLimitExceeded",
            "The maximum number of Tags for a resource has been reached.",
        )


class InvalidID(EC2ClientError):
    def __init__(self, resource_id: str):
        super().__init__("InvalidID", f"The ID '{resource_id}' is not valid")


class InvalidCIDRSubnetError(EC2ClientError):
    def __init__(self, cidr: Any):
        super().__init__(
            "InvalidParameterValue", f"invalid CIDR subnet specification: {cidr}"
        )


class RulesPerSecurityGroupLimitExceededError(EC2ClientError):
    def __init__(self) -> None:
        super().__init__(
            "RulesPerSecurityGroupLimitExceeded",
            "The maximum number of rules per security group " "has been reached.",
        )


class MotoNotImplementedError(NotImplementedError):
    def __init__(self, blurb: str):
        super().__init__(
            f"{blurb} has not been implemented in Moto yet."
            " Feel free to open an issue at"
            " https://github.com/getmoto/moto/issues"
        )


class FilterNotImplementedError(MotoNotImplementedError):
    def __init__(self, filter_name: str, method_name: Optional[str]):
        super().__init__(f"The filter '{filter_name}' for {method_name}")


class CidrLimitExceeded(EC2ClientError):
    def __init__(self, vpc_id: str, max_cidr_limit: int):
        super().__init__(
            "CidrLimitExceeded",
            f"This network '{vpc_id}' has met its maximum number of allowed CIDRs: {max_cidr_limit}",
        )


class UnsupportedTenancy(EC2ClientError):
    def __init__(self, tenancy: str):
        super().__init__(
            "UnsupportedTenancy", f"The tenancy value {tenancy} is not supported."
        )


class OperationNotPermitted(EC2ClientError):
    def __init__(self, association_id: str):
        super().__init__(
            "OperationNotPermitted",
            f"The vpc CIDR block with association ID {association_id} may not be disassociated. It is the primary IPv4 CIDR block of the VPC",
        )


class InvalidAvailabilityZoneError(EC2ClientError):
    def __init__(
        self, availability_zone_value: Optional[str], valid_availability_zones: str
    ):
        super().__init__(
            "InvalidParameterValue",
            f"Value ({availability_zone_value}) for parameter availabilityZone is invalid. "
            f"Subnets can currently only be created in the following availability zones: {valid_availability_zones}.",
        )


class AvailabilityZoneNotFromRegionError(EC2ClientError):
    def __init__(self, availability_zone_value: str):
        super().__init__(
            "InvalidParameterValue",
            f"Invalid Availability Zone ({availability_zone_value})",
        )


class NetworkAclEntryAlreadyExistsError(EC2ClientError):
    def __init__(self, rule_number: str):
        super().__init__(
            "NetworkAclEntryAlreadyExists",
            f"The network acl entry identified by {rule_number} already exists.",
        )


class InvalidSubnetRangeError(EC2ClientError):
    def __init__(self, cidr_block: str):
        super().__init__("InvalidSubnet.Range", f"The CIDR '{cidr_block}' is invalid.")


class InvalidCIDRBlockParameterError(EC2ClientError):
    def __init__(self, cidr_block: str):
        super().__init__(
            "InvalidParameterValue",
            f"Value ({cidr_block}) for parameter cidrBlock is invalid. This is not a valid CIDR block.",
        )


class InvalidDestinationCIDRBlockParameterError(EC2ClientError):
    def __init__(self, cidr_block: str):
        super().__init__(
            "InvalidParameterValue",
            f"Value ({cidr_block}) for parameter destinationCidrBlock is invalid. This is not a valid CIDR block.",
        )


class InvalidSubnetConflictError(EC2ClientError):
    def __init__(self, cidr_block: str):
        super().__init__(
            "InvalidSubnet.Conflict",
            f"The CIDR '{cidr_block}' conflicts with another subnet",
        )


class InvalidVPCRangeError(EC2ClientError):
    def __init__(self, cidr_block: str):
        super().__init__("InvalidVpc.Range", f"The CIDR '{cidr_block}' is invalid.")


# accept exception
class OperationNotPermitted2(EC2ClientError):
    def __init__(self, client_region: str, pcx_id: str, acceptor_region: str):
        super().__init__(
            "OperationNotPermitted",
            f"Incorrect region ({client_region}) specified for this request.VPC peering connection {pcx_id} must be accepted in region {acceptor_region}",
        )


# reject exception
class OperationNotPermitted3(EC2ClientError):
    def __init__(self, client_region: str, pcx_id: str, acceptor_region: str):
        super().__init__(
            "OperationNotPermitted",
            f"Incorrect region ({client_region}) specified for this request.VPC peering connection {pcx_id} must be accepted or rejected in region {acceptor_region}",
        )


class OperationNotPermitted4(EC2ClientError):
    def __init__(self, instance_id: str):
        super().__init__(
            "OperationNotPermitted",
            f"The instance '{instance_id}' may not be terminated. Modify its 'disableApiTermination' instance attribute and try again.",
        )


class InvalidLaunchTemplateNameAlreadyExistsError(EC2ClientError):
    def __init__(self) -> None:
        super().__init__(
            "InvalidLaunchTemplateName.AlreadyExistsException",
            "Launch template name already in use.",
        )


class InvalidLaunchTemplateNameNotFoundError(EC2ClientError):
    def __init__(self) -> None:
        super().__init__(
            "InvalidLaunchTemplateName.NotFoundException",
            "At least one of the launch templates specified in the request does not exist.",
        )


class InvalidLaunchTemplateNameNotFoundWithNameError(EC2ClientError):
    def __init__(self, name: str):
        super().__init__(
            "InvalidLaunchTemplateName.NotFoundException",
            f"The specified launch template, with template name {name}, does not exist",
        )


class InvalidParameterDependency(EC2ClientError):
    def __init__(self, param: str, param_needed: str):
        super().__init__(
            "InvalidParameterDependency",
            f"The parameter [{param}] requires the parameter {param_needed} to be set.",
        )


class IncorrectStateIamProfileAssociationError(EC2ClientError):
    def __init__(self, instance_id: str):
        super().__init__(
            "IncorrectState",
            f"There is an existing association for instance {instance_id}",
        )


class InvalidAssociationIDIamProfileAssociationError(EC2ClientError):
    def __init__(self, association_id: str):
        super().__init__(
            "InvalidAssociationID.NotFound",
            f"An invalid association-id of '{association_id}' was given",
        )


class InvalidVpcEndPointIdError(EC2ClientError):
    def __init__(self, vpc_end_point_id: str):
        super().__init__(
            "InvalidVpcEndpointId.NotFound",
            f"The VpcEndPoint ID '{vpc_end_point_id}' does not exist",
        )


class InvalidTaggableResourceType(EC2ClientError):
    def __init__(self, resource_type: str):
        super().__init__(
            "InvalidParameterValue",
            f"'{resource_type}' is not a valid taggable resource type for this operation.",
        )


class GenericInvalidParameterValueError(EC2ClientError):
    def __init__(self, attribute: str, value: str):
        super().__init__(
            "InvalidParameterValue",
            f"invalid value for parameter {attribute}: {value}",
        )


class InvalidSubnetCidrBlockAssociationID(EC2ClientError):
    def __init__(self, association_id: str):
        super().__init__(
            "InvalidSubnetCidrBlockAssociationID.NotFound",
            f"The subnet CIDR block with association ID '{association_id}' does not exist",
        )


class InvalidCarrierGatewayID(EC2ClientError):
    def __init__(self, carrier_gateway_id: str):
        super().__init__(
            "InvalidCarrierGatewayID.NotFound",
            f"The CarrierGateway ID '{carrier_gateway_id}' does not exist",
        )


class NoLoadBalancersProvided(EC2ClientError):
    def __init__(self) -> None:
        super().__init__(
            "InvalidParameter",
            "exactly one of network_load_balancer_arn or gateway_load_balancer_arn is a required member",
        )


class UnknownVpcEndpointService(EC2ClientError):
    def __init__(self, service_id: str):
        super().__init__(
            "InvalidVpcEndpointServiceId.NotFound",
            f"The VpcEndpointService Id '{service_id}' does not exist",
        )


class AuthFailureRestricted(RESTError):
    """Replicate real world issue https://github.com/aws/aws-cli/issues/1083"""

    code = 401

    def __init__(self) -> None:
        super().__init__(
            "AuthFailure",
            "Unauthorized attempt to access restricted resource",
        )



"""comprehend base URL and path."""
from .responses import ComprehendResponse

url_bases = [
    r"https?://comprehend\.(.+)\.amazonaws\.com",
]


url_paths = {
    "{0}/$": ComprehendResponse.dispatch,
}



import json

from moto.core.responses import BaseResponse
from moto.moto_api._internal import mock_random

from .models import ssoadmin_backends, SSOAdminBackend


class SSOAdminResponse(BaseResponse):
    """Handler for SSOAdmin requests and responses."""

    def __init__(self) -> None:
        super().__init__(service_name="sso-admin")

    @property
    def ssoadmin_backend(self) -> SSOAdminBackend:
        """Return backend instance specific for this region."""
        return ssoadmin_backends[self.current_account][self.region]

    def create_account_assignment(self) -> str:
        params = json.loads(self.body)
        instance_arn = params.get("InstanceArn")
        target_id = params.get("TargetId")
        target_type = params.get("TargetType")
        permission_set_arn = params.get("PermissionSetArn")
        principal_type = params.get("PrincipalType")
        principal_id = params.get("PrincipalId")
        summary = self.ssoadmin_backend.create_account_assignment(
            instance_arn=instance_arn,
            target_id=target_id,
            target_type=target_type,
            permission_set_arn=permission_set_arn,
            principal_type=principal_type,
            principal_id=principal_id,
        )
        summary["Status"] = "SUCCEEDED"
        summary["RequestId"] = str(mock_random.uuid4())
        return json.dumps({"AccountAssignmentCreationStatus": summary})

    def delete_account_assignment(self) -> str:
        params = json.loads(self.body)
        instance_arn = params.get("InstanceArn")
        target_id = params.get("TargetId")
        target_type = params.get("TargetType")
        permission_set_arn = params.get("PermissionSetArn")
        principal_type = params.get("PrincipalType")
        principal_id = params.get("PrincipalId")
        summary = self.ssoadmin_backend.delete_account_assignment(
            instance_arn=instance_arn,
            target_id=target_id,
            target_type=target_type,
            permission_set_arn=permission_set_arn,
            principal_type=principal_type,
            principal_id=principal_id,
        )
        summary["Status"] = "SUCCEEDED"
        summary["RequestId"] = str(mock_random.uuid4())
        return json.dumps({"AccountAssignmentDeletionStatus": summary})

    def list_account_assignments(self) -> str:
        params = json.loads(self.body)
        instance_arn = params.get("InstanceArn")
        account_id = params.get("AccountId")
        permission_set_arn = params.get("PermissionSetArn")
        assignments = self.ssoadmin_backend.list_account_assignments(
            instance_arn=instance_arn,
            account_id=account_id,
            permission_set_arn=permission_set_arn,
        )
        return json.dumps({"AccountAssignments": assignments})

    def create_permission_set(self) -> str:
        name = self._get_param("Name")
        description = self._get_param("Description")
        instance_arn = self._get_param("InstanceArn")
        session_duration = self._get_param("SessionDuration", 3600)
        relay_state = self._get_param("RelayState")
        tags = self._get_param("Tags")

        permission_set = self.ssoadmin_backend.create_permission_set(
            name=name,
            description=description,
            instance_arn=instance_arn,
            session_duration=session_duration,
            relay_state=relay_state,
            tags=tags,
        )

        return json.dumps({"PermissionSet": permission_set})

    def delete_permission_set(self) -> str:
        params = json.loads(self.body)
        instance_arn = params.get("InstanceArn")
        permission_set_arn = params.get("PermissionSetArn")
        self.ssoadmin_backend.delete_permission_set(
            instance_arn=instance_arn,
            permission_set_arn=permission_set_arn,
        )
        return "{}"

    def update_permission_set(self) -> str:
        instance_arn = self._get_param("InstanceArn")
        permission_set_arn = self._get_param("PermissionSetArn")
        description = self._get_param("Description")
        session_duration = self._get_param("SessionDuration", 3600)
        relay_state = self._get_param("RelayState")

        self.ssoadmin_backend.update_permission_set(
            instance_arn=instance_arn,
            permission_set_arn=permission_set_arn,
            description=description,
            session_duration=session_duration,
            relay_state=relay_state,
        )
        return "{}"

    def describe_permission_set(self) -> str:
        instance_arn = self._get_param("InstanceArn")
        permission_set_arn = self._get_param("PermissionSetArn")

        permission_set = self.ssoadmin_backend.describe_permission_set(
            instance_arn=instance_arn,
            permission_set_arn=permission_set_arn,
        )
        return json.dumps({"PermissionSet": permission_set})

    def list_permission_sets(self) -> str:
        instance_arn = self._get_param("InstanceArn")
        max_results = self._get_int_param("MaxResults")
        next_token = self._get_param("NextToken")
        permission_sets, next_token = self.ssoadmin_backend.list_permission_sets(
            instance_arn=instance_arn, max_results=max_results, next_token=next_token
        )
        permission_set_ids = []
        for permission_set in permission_sets:
            permission_set_ids.append(permission_set.permission_set_arn)
        response = {"PermissionSets": permission_set_ids}
        if next_token:
            response["NextToken"] = next_token
        return json.dumps(response)



"""Pagination control model for Route53Resolver."""

from .exceptions import InvalidNextTokenException

PAGINATION_MODEL = {
    "list_resolver_endpoint_ip_addresses": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 100,
        "unique_attribute": "IpId",
    },
    "list_resolver_endpoints": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 100,
        "unique_attribute": "id",
        "fail_on_invalid_token": InvalidNextTokenException,
    },
    "list_resolver_rules": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 100,
        "unique_attribute": "id",
    },
    "list_resolver_rule_associations": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 100,
        "unique_attribute": "id",
    },
    "list_tags_for_resource": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 100,
        "unique_attribute": "Key",
        "fail_on_invalid_token": InvalidNextTokenException,
    },
}



"""apigatewaymanagementapi base URL and path."""
from .responses import ApiGatewayManagementApiResponse

url_bases = [r"https?://execute-api\.(.+)\.amazonaws\.com"]


response = ApiGatewayManagementApiResponse()


url_paths = {
    "{0}/@connections/(?P<connection_id>[^/]+)$": response.dispatch,
}



from ._base_response import EC2BaseResponse


class HostsResponse(EC2BaseResponse):
    def allocate_hosts(self) -> str:
        params = self._get_params()
        quantity = int(params.get("Quantity"))  # type: ignore
        host_recovery = params.get("HostRecovery")
        zone = params.get("AvailabilityZone")
        instance_type = params.get("InstanceType")
        instance_family = params.get("InstanceFamily")
        auto_placement = params.get("AutoPlacement")
        tags = self._parse_tag_specification()
        host_tags = tags.get("dedicated-host", {})
        host_ids = self.ec2_backend.allocate_hosts(
            quantity,
            host_recovery,
            zone,
            instance_type,
            instance_family,
            auto_placement,
            host_tags,
        )
        template = self.response_template(EC2_ALLOCATE_HOSTS)
        return template.render(host_ids=host_ids)

    def describe_hosts(self) -> str:
        host_ids = list(self._get_params().get("HostId", {}).values())
        filters = self._filters_from_querystring()
        hosts = self.ec2_backend.describe_hosts(host_ids, filters)
        template = self.response_template(EC2_DESCRIBE_HOSTS)
        return template.render(account_id=self.current_account, hosts=hosts)

    def modify_hosts(self) -> str:
        params = self._get_params()
        host_ids = list(self._get_params().get("HostId", {}).values())
        auto_placement = params.get("AutoPlacement")
        host_recovery = params.get("HostRecovery")
        instance_type = params.get("InstanceType")
        instance_family = params.get("InstanceFamily")
        self.ec2_backend.modify_hosts(
            host_ids, auto_placement, host_recovery, instance_type, instance_family
        )
        template = self.response_template(EC2_MODIFY_HOSTS)
        return template.render(host_ids=host_ids)

    def release_hosts(self) -> str:
        host_ids = list(self._get_params().get("HostId", {}).values())
        self.ec2_backend.release_hosts(host_ids)
        template = self.response_template(EC2_RELEASE_HOSTS)
        return template.render(host_ids=host_ids)


EC2_ALLOCATE_HOSTS = """<AllocateHostsResult xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
<requestId>fdcdcab1-ae5c-489e-9c33-4637c5dda355</requestId>
    <hostIdSet>
        {% for host_id in host_ids %}
            <item>{{ host_id }}</item>
        {% endfor %}
    </hostIdSet>
</AllocateHostsResult>"""


EC2_DESCRIBE_HOSTS = """<DescribeHostsResult xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
<requestId>fdcdcab1-ae5c-489e-9c33-4637c5dda355</requestId>
    <hostSet>
        {% for host in hosts %}
            <item>
                <autoPlacement>{{ host.auto_placement }}</autoPlacement>
                <availabilityZone>{{ host.zone }}</availabilityZone>
                <availableCapacity></availableCapacity>
                <hostId>{{ host.id }}</hostId>
                <state>{{ host.state }}</state>
                <hostProperties>
                  {% if host.instance_type %}
                    <instanceType>{{ host.instance_type }}</instanceType>
                  {% endif %}
                  {% if host.instance_family %}
                    <instanceFamily>{{ host.instance_family }}</instanceFamily>
                  {% endif %}
                </hostProperties>
                <hostReservationId>reserv_id</hostReservationId>
                <instances>
                </instances>
                <ownerId>{{ account_id }}</ownerId>
                <hostRecovery>{{ host.host_recovery }}</hostRecovery>
                <tagSet>
                  {% for tag in host.get_tags() %}
                    <item>
                      <key>{{ tag.key }}</key>
                      <value>{{ tag.value }}</value>
                    </item>
                  {% endfor %}
                </tagSet>
            </item>
        {% endfor %}
    </hostSet>
</DescribeHostsResult>"""


EC2_MODIFY_HOSTS = """<ModifyHostsResult xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
<requestId>fdcdcab1-ae5c-489e-9c33-4637c5dda355</requestId>
    <successful>
        {% for host_id in host_ids %}
            <item>{{ host_id }}</item>
        {% endfor %}
    </successful>
</ModifyHostsResult>"""


EC2_RELEASE_HOSTS = """<ReleaseHostsResult xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
<requestId>fdcdcab1-ae5c-489e-9c33-4637c5dda355</requestId>
    <successful>
        {% for host_id in host_ids %}
            <item>{{ host_id }}</item>
        {% endfor %}
    </successful>
</ReleaseHostsResult>"""



"""Handles incoming ebs requests, invokes methods, returns responses."""
import json
from typing import Any

from moto.core.responses import BaseResponse
from moto.core.common_types import TYPE_RESPONSE
from .models import ebs_backends, EBSBackend


class EBSResponse(BaseResponse):
    """Handler for EBS requests and responses."""

    def __init__(self) -> None:
        super().__init__(service_name="ebs")

    @property
    def ebs_backend(self) -> EBSBackend:
        """Return backend instance specific for this region."""
        return ebs_backends[self.current_account][self.region]

    def snapshots(self, request: Any, full_url: str, headers: Any) -> TYPE_RESPONSE:  # type: ignore[return]
        self.setup_class(request, full_url, headers)
        if request.method == "POST":
            return self.start_snapshot()

    def snapshot_block(self, request: Any, full_url: str, headers: Any) -> TYPE_RESPONSE:  # type: ignore[return]
        self.setup_class(request, full_url, headers)
        if request.method == "PUT":
            return self.put_snapshot_block(full_url, headers)
        if request.method == "GET":
            return self.get_snapshot_block()

    def snapshot_blocks(self, request: Any, full_url: str, headers: Any) -> TYPE_RESPONSE:  # type: ignore[return]
        self.setup_class(request, full_url, headers)
        if request.method == "GET":
            return self.list_snapshot_blocks()

    def start_snapshot(self) -> TYPE_RESPONSE:
        """
        The following parameters are not yet implemented: ParentSnapshotId, ClientToken, Encrypted, KmsKeyArn, Timeout
        """
        params = json.loads(self.body)
        volume_size = params.get("VolumeSize")
        tags = params.get("Tags")
        description = params.get("Description")
        snapshot = self.ebs_backend.start_snapshot(
            volume_size=volume_size,
            tags=tags,
            description=description,
        )
        return 200, {}, json.dumps(snapshot.to_json())

    def complete_snapshot(
        self, request: Any, full_url: str, headers: Any
    ) -> TYPE_RESPONSE:
        """
        The following parameters are not yet supported: ChangedBlocksCount, Checksum, ChecksumAlgorithm, ChecksumAggregationMethod
        """
        self.setup_class(request, full_url, headers)
        snapshot_id = full_url.split("/")[-1]
        status = self.ebs_backend.complete_snapshot(snapshot_id=snapshot_id)
        return 200, {}, json.dumps(status)

    def put_snapshot_block(self, full_url: str, headers: Any) -> TYPE_RESPONSE:
        """
        The following parameters are currently not taken into account: DataLength, Progress.
        The Checksum and ChecksumAlgorithm are taken at face-value, but no validation takes place.
        """
        snapshot_id = full_url.split("/")[-3]
        block_index = full_url.split("/")[-1]
        block_data = self.body
        headers = {k.lower(): v for k, v in headers.items()}
        checksum = headers.get("x-amz-checksum")
        checksum_algorithm = headers.get("x-amz-checksum-algorithm")
        data_length = headers.get("x-amz-data-length")
        checksum, checksum_algorithm = self.ebs_backend.put_snapshot_block(
            snapshot_id=snapshot_id,
            block_index=block_index,
            block_data=block_data,
            checksum=checksum,
            checksum_algorithm=checksum_algorithm,
            data_length=data_length,
        )
        return (
            200,
            {
                "x-amz-Checksum": checksum,
                "x-amz-Checksum-Algorithm": checksum_algorithm,
            },
            "{}",
        )

    def get_snapshot_block(self) -> TYPE_RESPONSE:
        snapshot_id = self.path.split("/")[-3]
        block_index = self.path.split("/")[-1]
        block = self.ebs_backend.get_snapshot_block(
            snapshot_id=snapshot_id,
            block_index=block_index,
        )
        headers = {
            "x-amz-Checksum": block.checksum,
            "x-amz-Checksum-Algorithm": block.checksum_algorithm,
            "x-amz-Data-Length": block.data_length,
        }
        return 200, headers, block.block_data

    def snapshot_changed_blocks(
        self, request: Any, full_url: str, headers: Any
    ) -> TYPE_RESPONSE:
        self.setup_class(request, full_url, headers)
        first_snapshot_id = self._get_params().get("firstSnapshotId")
        second_snapshot_id = self.path.split("/")[-2]
        changed_blocks, snapshot = self.ebs_backend.list_changed_blocks(
            first_snapshot_id=first_snapshot_id,  # type: ignore[arg-type]
            second_snapshot_id=second_snapshot_id,
        )
        blocks = [
            {"BlockIndex": idx, "FirstBlockToken": x, "SecondBlockToken": y}
            for idx, (x, y) in changed_blocks.items()
        ]
        return (
            200,
            {},
            json.dumps(
                dict(
                    ChangedBlocks=blocks,
                    VolumeSize=snapshot.volume_size,
                    BlockSize=snapshot.block_size,
                )
            ),
        )

    def list_snapshot_blocks(self) -> TYPE_RESPONSE:
        """
        The following parameters are not yet implemented: NextToken, MaxResults, StartingBlockIndex
        """
        snapshot_id = self.path.split("/")[-2]
        snapshot = self.ebs_backend.list_snapshot_blocks(
            snapshot_id=snapshot_id,
        )
        blocks = [
            {"BlockIndex": idx, "BlockToken": b.block_token}
            for idx, b in snapshot.blocks.items()
        ]
        return (
            200,
            {},
            json.dumps(
                dict(
                    Blocks=blocks,
                    VolumeSize=snapshot.volume_size,
                    BlockSize=snapshot.block_size,
                )
            ),
        )



from moto.utilities.distutils_version import LooseVersion

try:
    from importlib.metadata import version
except ImportError:
    from importlib_metadata import version  # type: ignore[no-redef]


RESPONSES_VERSION = version("responses")
WERKZEUG_VERSION = version("werkzeug")


def is_responses_0_17_x() -> bool:
    return LooseVersion(RESPONSES_VERSION) >= LooseVersion("0.17.0")


def is_werkzeug_2_3_x() -> bool:
    return LooseVersion(WERKZEUG_VERSION) >= LooseVersion("2.3.0")



from typing import Any, Dict, List, Optional
from moto.core import BaseBackend, BackendDict, BaseModel
from moto.moto_api._internal import mock_random
from datetime import datetime

from .exceptions import DetectorNotFoundException, FilterNotFoundException


class GuardDutyBackend(BaseBackend):
    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.admin_account_ids: List[str] = []
        self.detectors: Dict[str, Detector] = {}

    def create_detector(
        self,
        enable: bool,
        finding_publishing_frequency: str,
        data_sources: Dict[str, Any],
        tags: Dict[str, str],
    ) -> str:
        if finding_publishing_frequency not in [
            "FIFTEEN_MINUTES",
            "ONE_HOUR",
            "SIX_HOURS",
        ]:
            finding_publishing_frequency = "SIX_HOURS"

        detector = Detector(
            account_id=self.account_id,
            created_at=datetime.now(),
            finding_publish_freq=finding_publishing_frequency,
            enabled=enable,
            datasources=data_sources,
            tags=tags,
        )
        self.detectors[detector.id] = detector
        return detector.id

    def create_filter(
        self,
        detector_id: str,
        name: str,
        action: str,
        description: str,
        finding_criteria: Dict[str, Any],
        rank: int,
    ) -> None:
        detector = self.get_detector(detector_id)
        _filter = Filter(name, action, description, finding_criteria, rank)
        detector.add_filter(_filter)

    def delete_detector(self, detector_id: str) -> None:
        self.detectors.pop(detector_id, None)

    def delete_filter(self, detector_id: str, filter_name: str) -> None:
        detector = self.get_detector(detector_id)
        detector.delete_filter(filter_name)

    def enable_organization_admin_account(self, admin_account_id: str) -> None:
        self.admin_account_ids.append(admin_account_id)

    def list_organization_admin_accounts(self) -> List[str]:
        """
        Pagination is not yet implemented
        """
        return self.admin_account_ids

    def list_detectors(self) -> List[str]:
        """
        The MaxResults and NextToken-parameter have not yet been implemented.
        """
        detectorids = []
        for detector in self.detectors:
            detectorids.append(self.detectors[detector].id)
        return detectorids

    def get_detector(self, detector_id: str) -> "Detector":
        if detector_id not in self.detectors:
            raise DetectorNotFoundException
        return self.detectors[detector_id]

    def get_filter(self, detector_id: str, filter_name: str) -> "Filter":
        detector = self.get_detector(detector_id)
        return detector.get_filter(filter_name)

    def update_detector(
        self,
        detector_id: str,
        enable: bool,
        finding_publishing_frequency: str,
        data_sources: Dict[str, Any],
    ) -> None:
        detector = self.get_detector(detector_id)
        detector.update(enable, finding_publishing_frequency, data_sources)

    def update_filter(
        self,
        detector_id: str,
        filter_name: str,
        action: str,
        description: str,
        finding_criteria: Dict[str, Any],
        rank: int,
    ) -> None:
        detector = self.get_detector(detector_id)
        detector.update_filter(
            filter_name,
            action=action,
            description=description,
            finding_criteria=finding_criteria,
            rank=rank,
        )


class Filter(BaseModel):
    def __init__(
        self,
        name: str,
        action: str,
        description: str,
        finding_criteria: Dict[str, Any],
        rank: int,
    ):
        self.name = name
        self.action = action
        self.description = description
        self.finding_criteria = finding_criteria
        self.rank = rank or 1

    def update(
        self,
        action: Optional[str],
        description: Optional[str],
        finding_criteria: Optional[Dict[str, Any]],
        rank: Optional[int],
    ) -> None:
        if action is not None:
            self.action = action
        if description is not None:
            self.description = description
        if finding_criteria is not None:
            self.finding_criteria = finding_criteria
        if rank is not None:
            self.rank = rank

    def to_json(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "action": self.action,
            "description": self.description,
            "findingCriteria": self.finding_criteria,
            "rank": self.rank,
        }


class Detector(BaseModel):
    def __init__(
        self,
        account_id: str,
        created_at: datetime,
        finding_publish_freq: str,
        enabled: bool,
        datasources: Dict[str, Any],
        tags: Dict[str, str],
    ):
        self.id = mock_random.get_random_hex(length=32)
        self.created_at = created_at
        self.finding_publish_freq = finding_publish_freq
        self.service_role = f"arn:aws:iam::{account_id}:role/aws-service-role/guardduty.amazonaws.com/AWSServiceRoleForAmazonGuardDuty"
        self.enabled = enabled
        self.updated_at = created_at
        self.datasources = datasources or {}
        self.tags = tags or {}

        self.filters: Dict[str, Filter] = dict()

    def add_filter(self, _filter: Filter) -> None:
        self.filters[_filter.name] = _filter

    def delete_filter(self, filter_name: str) -> None:
        self.filters.pop(filter_name, None)

    def get_filter(self, filter_name: str) -> Filter:
        if filter_name not in self.filters:
            raise FilterNotFoundException
        return self.filters[filter_name]

    def update_filter(
        self,
        filter_name: str,
        action: str,
        description: str,
        finding_criteria: Dict[str, Any],
        rank: int,
    ) -> None:
        _filter = self.get_filter(filter_name)
        _filter.update(
            action=action,
            description=description,
            finding_criteria=finding_criteria,
            rank=rank,
        )

    def update(
        self,
        enable: bool,
        finding_publishing_frequency: str,
        data_sources: Dict[str, Any],
    ) -> None:
        if enable is not None:
            self.enabled = enable
        if finding_publishing_frequency is not None:
            self.finding_publish_freq = finding_publishing_frequency
        if data_sources is not None:
            self.datasources = data_sources

    def to_json(self) -> Dict[str, Any]:
        data_sources = {
            "cloudTrail": {"status": "DISABLED"},
            "dnsLogs": {"status": "DISABLED"},
            "flowLogs": {"status": "DISABLED"},
            "s3Logs": {
                "status": "ENABLED"
                if (self.datasources.get("s3Logs") or {}).get("enable")
                else "DISABLED"
            },
            "kubernetes": {
                "auditLogs": {
                    "status": "ENABLED"
                    if self.datasources.get("kubernetes", {})
                    .get("auditLogs", {})
                    .get("enable")
                    else "DISABLED"
                }
            },
        }
        return {
            "createdAt": self.created_at.strftime("%Y-%m-%dT%H:%M:%S.%fZ"),
            "findingPublishingFrequency": self.finding_publish_freq,
            "serviceRole": self.service_role,
            "status": "ENABLED" if self.enabled else "DISABLED",
            "updatedAt": self.updated_at.strftime("%Y-%m-%dT%H:%M:%S.%fZ"),
            "dataSources": data_sources,
            "tags": self.tags,
        }


guardduty_backends = BackendDict(GuardDutyBackend, "guardduty")
